<div class="modal modal-blur fade" id="modal-delete" tabindex="-1" role="dialog" aria-hidden="true">
  <div class="modal-dialog modal-sm modal-dialog-centered" role="document">
    <div class="modal-content">
      <button type="button" class="btn-close" data-bs-dismiss="modal" aria-label="Close"></button>
      <div class="modal-status bg-danger"></div>
      <div class="modal-body text-center py-4">
        <i class="text-danger fa-solid fa-triangle-exclamation fa-2x"></i>
        <h3>Please confirm</h3>
        <div id="modal-delete-text" class="text-muted"></div>
      </div>
      <div class="modal-footer">
        <div class="w-100">
          <div class="row">
            <div class="col">
              <a href="#" class="btn w-100" data-bs-dismiss="modal">
                Cancel
              </a>
            </div>
            <div class="col">
              <button id="modal-delete-button" data-url="" class="btn btn-danger w-100">
                Delete
              </button>
            </div>
          </div>
        </div>
      </div>
    </div>
  </div>
</div>
